#ifdef __APPLE__
#include <GLUT/glut.h>    // Header File For The GLUT Library 
#include <OpenGL/OpenGL.h>    // Header File For OpenGL
#else
#include <GL/glut.h>    // Header File For The GLUT Library 
#endif
#include <stdio.h>
#include <unistd.h>     // Header File For sleeping.
#include <stdlib.h>

#include <string.h>
/* ASCII code for the escape key. */
#define ESCAPE 27

/* Fullscreen State */

/* Are we entering fullscreen or leaving it. */
GLboolean fullscreen = GL_TRUE;

/* The number of our GLUT window */
int window; 

const char model_file_name[] = "./mushroom.geo";

typedef struct _vertex_t {
  GLfloat x,y,z;
} vertex_t;

typedef struct _color_t {
  GLfloat r,g,b;
} color_t;

color_t possible_colors[] = {\
  {.r = 1.0, .g = 0.0, .b = 0.0},\
  {.r = 0.0, .g = 0.0, .b = 1.0},\
  {.r = 0.0, .g = 1.0, .b = 1.0},\
  {.r = 0.0, .g = 1.0, .b = 0.0},\
  {.r = 1.0, .g = 1.0, .b = 0.0},\
  {.r = 1.0, .g = 0.0, .b = 1.0},\
};

GLushort poly_count    = 0;    // Number of polygons.
GLushort index_count   = 0;    // number of indicies
GLushort vertex_count  = 0;    // number of verticies.
GLushort *polygons     = NULL; // array of the sizes of each polygon
GLushort *index_array  = NULL; // array of indicies into the virtex array
vertex_t *vertex_array = NULL; // array of verticies (length 3*max(index_array)
color_t  *color_array  = NULL; // array of verticies (length 3*max(index_array)
vertex_t min_vertex, max_vertex, vertex_center;

/* Mouse and camera variables */
GLboolean mouseLeftDown  = GL_FALSE;
GLboolean mouseRightDown = GL_FALSE;
GLboolean loading_model  = GL_FALSE;
GLboolean wire_frame = GL_FALSE;
float mouseX, mouseY;
float cameraAngleX;
float cameraAngleY;
float cameraDistance;

GLuint vertexVBO, indexVBO, colorVBO;// model VBO and index VBO
/* parsing prototypes */
int ignore_comments(FILE *modelFile); 
void read_geo_model_to_vbo(const char *file_name);
void create_vbos();
void destroy_vbos();

/* Window Control Prototypes */
void ToggleFullScreen();
void keyPressed(unsigned char key, int x, int y);
/* GLUT prototypes */
void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ);
void InitGL(int Width, int Height);
void mouseCB(int button, int stat, int x, int y);
void mouseMotionCB(int x, int y);
void ReSizeGLScene(int Width, int Height);
void DrawGLScene();
void initLights();
GLboolean isExtensionSupported(const char *extension); 

/* VBO prototypes */
GLuint createVBO(void* data, int dataSize, GLenum target, GLenum usage);
void deleteVBO(GLuint vboId);

/* main entery point of the program */

int main(int argc, char **argv) {


  glutInit(&argc, argv);  
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);  
  /* get a 640 x 480 window */
  glutInitWindowSize(640, 480);  
  /* the window starts at the upper left corner of the screen */
  glutInitWindowPosition(0, 0);  
  /* Open a window */  
  window = glutCreateWindow("James Harris's GL VBO program. 2011");  
  /* Register the function to do all our OpenGL drawing. */
  glutDisplayFunc(&DrawGLScene);  
  /* Go fullscreen.  This is the soonest we could possibly go fullscreen. */
  ToggleFullScreen();
  /* Register the function called when our window is resized. */
  glutReshapeFunc(&ReSizeGLScene);
  /* Register the function called when the keyboard is pressed. */
  glutKeyboardFunc(&keyPressed);
  glutMouseFunc(&mouseCB);
  glutMotionFunc(&mouseMotionCB);
  /* Initialize our window. */
  InitGL(640, 480);

  /* Buld our VBOS from the file */
  if (!isExtensionSupported("GL_ARB_vertex_buffer_object")) {
    fprintf(stderr, "VBOs Not supported on this hardware! Quitting!\n");
    exit(-1);
  }
  // Populate Model Matricies.
  read_geo_model_to_vbo(model_file_name);
  /* read_geo_model_to_vbo("./mushroom.geo"); */
  /* Start Event Processing Engine */  
  glutMainLoop();  
  return 1;
}

// Checks for a comment (line beginning with '#') and skips the line).
// Returns GL_TRUE if comment and GL_FALSE if it is not. 
int ignore_comments(FILE *modelFile) {
  char start_of_line = getc(modelFile);
  char curr_char = start_of_line;
  // quit if were reach the end of the file pre-maturely.
  if (feof(modelFile)) {
    fprintf(stderr, "File format error, hit EOF before finishing the input file.\n");
    return 0;
  }
  // while lines are only comments.
  while(start_of_line == '#') {
    while(curr_char != '\n') {
      curr_char = getc(modelFile); //ignore the current line
    }
    start_of_line = getc(modelFile);
    curr_char = start_of_line;
  } 
  ungetc(start_of_line, modelFile); // back up one character.
  return 1; // tell caller to execute the state machine.
}

/* Read the model File. */
void read_geo_model_to_vbo(const char *file_name) {
  GLushort current_index = 0;
  GLushort current_polygon = 0;
  GLushort current_vertex = 0;
  GLushort current_color = 0;
  unsigned short number_read = 0;
  GLfloat x, y, z, h; // x y z corrdinates and h (homogeneous coordinates.)

  min_vertex.x  = 0;
  min_vertex.y  = 0;
  min_vertex.z  = 0;
  max_vertex.x  = 0;
  max_vertex.y  = 0;
  max_vertex.z  = 0;
  vertex_center.x  = 0;
  vertex_center.y  = 0;
  vertex_center.z  = 0;

  loading_model = GL_TRUE;

  FILE *model_file = fopen(file_name, "r");

  if(model_file != NULL) {

    ignore_comments(model_file);

    fscanf(model_file, "%hu\n", (unsigned short *)&poly_count);

    ignore_comments(model_file);

    polygons = (GLushort *)malloc(sizeof(GLushort) * poly_count);

    if (polygons == NULL) {
      fprintf(stderr, "Unable to allocate memory for the Polygons\n");
      exit(-1);
    }

    for (current_polygon = 0; current_polygon < poly_count; current_polygon++) {
      fscanf(model_file, "%hu\n", &number_read);
      polygons[current_polygon] = (GLushort)number_read;
      index_count += number_read;
    }

    ignore_comments(model_file);

    index_array = (GLushort *)malloc(sizeof(GLushort) * index_count);

    if (index_array == NULL) {
      fprintf(stderr, "Unable to allocate memory for the index_array.\n");
      exit(-1);
    }

    for (current_index = 0; current_index < index_count; current_index++) {
      fscanf(model_file, "%hu\n", &number_read);
      index_array[current_index] = (GLushort)number_read;
      vertex_count = vertex_count < number_read ? number_read : vertex_count;
    }

    ignore_comments(model_file);
    vertex_count++;

    vertex_array = (vertex_t *)malloc(sizeof(vertex_t) * vertex_count);

    if (vertex_array == NULL) {
      fprintf(stderr, "Unable to allocate memory for the vertex array.\n");
      exit(-1);
    }

    color_array = (color_t *)malloc(sizeof(color_t) * vertex_count);

    if (vertex_array == NULL) {
      fprintf(stderr, "Unable to allocate memory for the vertex array.\n");
      exit(-1);
    }


    for (current_vertex = 0; current_vertex < vertex_count; current_vertex++) {
      fscanf(model_file,"%f %f %f %f\n", &x, &y, &z, &h);

      /* Determine our range of verticies */
      min_vertex.x = min_vertex.x > x ? x : min_vertex.x ;
      min_vertex.y = min_vertex.y > y ? y : min_vertex.y ;
      min_vertex.z = min_vertex.z > z ? z : min_vertex.z ;
      /* Determine our range of verticies */
      max_vertex.x = max_vertex.x < x ? x : max_vertex.x ;
      max_vertex.y = max_vertex.y < y ? y : max_vertex.y ;
      max_vertex.z = max_vertex.z < z ? z : max_vertex.z ;

      vertex_count = vertex_count < number_read ? number_read : vertex_count;
      vertex_array[current_vertex].x = x;
      vertex_array[current_vertex].y = y;
      vertex_array[current_vertex].z = z;
      color_array[current_vertex] = possible_colors[current_color];
      current_color = current_color == 5 ? 0:current_color+1;
    }
    vertex_center.x = (min_vertex.x + max_vertex.x)/2;
    vertex_center.y = (min_vertex.y + max_vertex.y)/2;
    vertex_center.z = (min_vertex.z + max_vertex.z)/2;
    fclose(model_file);
    create_vbos();
  }

  loading_model = GL_FALSE;
}

/* Print the model to the shield */
void print_model() {
  int i;
  printf("Poly Count: %d\n", poly_count);
  printf("Polygons:\n");
  for (i = 0; i < poly_count; i++) {
    printf("%d\n", polygons[i]);
  }
  printf("Index Count: %d\n", index_count);
  printf("Indicies:\n");
  for (i = 0; i < index_count; i++) {
    printf("%d\n", index_array[i]);
  }
  printf("Vertex Count: %d\n", vertex_count);
  printf("Verticies:\n");
  for (i = 0; i < vertex_count; i++) {
    printf("x:%f, y:%f, z:%f, r:%f, g:%f, b:%f,\n", vertex_array[i].x,\
        vertex_array[i].y,\
        vertex_array[i].z,\
        color_array[i].r,\
        color_array[i].g,\
        color_array[i].b);
  }
}

/* Add function for toggling fullscreen mode on and off.*/
void ToggleFullScreen() {
  fullscreen = !fullscreen;

  if (fullscreen)
  {
    glutFullScreen();
  } else {
    /* Enter Windowed mode. */
    glutReshapeWindow(640, 480);
    /* The 30 pixel offset is for the Menu bar in os x. */
    glutPositionWindow(0,30);
  }
}

/* A general OpenGL initialization function.  Sets all of the initial parameters. */
void InitGL(int Width, int Height) {
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);		// This Will Clear The Background Color To Black
  glClearDepth(1.0);				// Enables Clearing Of The Depth Buffer
  glDepthFunc(GL_LESS);				// The Type Of Depth Test To Do
  glEnable(GL_DEPTH_TEST);			// Enables Depth Testing
  glShadeModel(GL_SMOOTH);			// Enables Smooth Color Shading

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();				// Reset The Projection Matrix

  gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,100.0f);	// Calculate The Aspect Ratio Of The Window
}

/* The function called when our window is resized (which shouldn't happen, because we're fullscreen) */
void ReSizeGLScene(int Width, int Height) {
  if (Height==0)				// Prevent A Divide By Zero If The Window Is Too Small
    Height=1;

  glViewport(0, 0, Width, Height);		// Reset The Current Viewport And Perspective Transformation

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,100.0f);
  glMatrixMode(GL_MODELVIEW);
}

/* The main drawing function. */
void DrawGLScene() {
  int i;
  if (!loading_model) {
    GLushort current_index = 0;
    GLushort count = 0;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear The Screen And The Depth Buffer
    setCamera(0,0,20,0,0,0); //reset the view and place the camera.

    // save the initial ModelView matrix before modifying ModelView matrix
    glPushMatrix();

    // tramsform camera
    glTranslatef(0, 0, cameraDistance);
    glRotatef(cameraAngleX, 1, 0, 0);   // pitch
    glRotatef(cameraAngleY, 0, 1, 0);   // heading

    glBindBuffer(GL_ARRAY_BUFFER, vertexVBO);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, sizeof(vertex_t), 0);

    glBindBuffer(GL_ARRAY_BUFFER, colorVBO);
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3, GL_FLOAT, sizeof(color_t), 0);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVBO);
    if (wire_frame) {
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    } else  {
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
    glPushMatrix();
    glTranslatef(-1* vertex_center.x,-1* vertex_center.y,-1* vertex_center.z);
    for (i = 0; i < poly_count; i++) {
      count = polygons[i];
      glDrawElements(GL_TRIANGLE_FAN, count, GL_UNSIGNED_SHORT, (GLushort*)0+current_index);
      current_index += count;
    }
    glPopMatrix(); // Remove Centering Transoformation.

    // swap the buffers to display, since double buffering is used.
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glPopMatrix();
    glutSwapBuffers();
  }
}

/* The function called whenever a key is pressed. */
void keyPressed(unsigned char key, int x, int y) {
  /* sleep to avoid thrashing this procedure */
  usleep(100);

  /* If the F key is pressed toggle full screen. */
  if (key == 102) {
    ToggleFullScreen();
    glutPostRedisplay();
  }

  if (!loading_model) {
    if (key == 109) {
      destroy_vbos();
      cameraAngleX = 0;
      cameraAngleY = 0;
      cameraDistance = 0;
      mouseX = 0;
      mouseY = 0;
      read_geo_model_to_vbo("mushroom.geo");
      glutPostRedisplay();
    }
    if (key == 104) {
      destroy_vbos();
      cameraAngleX = 0;
      cameraAngleY = 0;
      cameraDistance = 0;
      mouseX = 0;
      mouseY = 0;
      read_geo_model_to_vbo("scenefile1.geo");
      glutPostRedisplay();
    }
  }

  if (key == 119) {
    wire_frame = !wire_frame;
  }


  /* If escape is pressed, kill everything. */
  if (key == ESCAPE) 
  { 
    /* shut down our window */
    glutDestroyWindow(window); 

    /* exit the program...normal termination. */
    exit(0);                   
  }
}

/* build the VBOs */
void create_vbos() {
  vertexVBO = createVBO(vertex_array, sizeof(vertex_t)*vertex_count, GL_ARRAY_BUFFER, GL_STATIC_DRAW);
  colorVBO = createVBO(color_array, sizeof(color_t)*vertex_count, GL_ARRAY_BUFFER, GL_STATIC_DRAW);
  indexVBO = createVBO(index_array, sizeof(GLshort)*index_count, GL_ELEMENT_ARRAY_BUFFER, GL_STATIC_DRAW);
  free(vertex_array);
  free( color_array);
  free( index_array);
}

void destroy_vbos() {
  deleteVBO(vertexVBO);
  deleteVBO(colorVBO );
  deleteVBO(indexVBO );
}

/* From SongHo.ca tutorial */
/********************************************************************************
 * generate vertex buffer object and bind it with its data                      *
 * You must give 2 hints about data usage; target and mode, so that OpenGL can  *
 * decide which data should be stored and its location.                         *
 * VBO works with 2 different targets; GL_ARRAY_BUFFER_ARB for vertex arrays    *
 * and GL_ELEMENT_ARRAY_BUFFER_ARB for index array in glDrawElements().         *
 * The default target is GL_ARRAY_BUFFER_ARB.                                   *
 * By default, usage mode is set as GL_STATIC_DRAW_ARB.                         *
 * Other usages are GL_STREAM_DRAW_ARB, GL_STREAM_READ_ARB, GL_STREAM_COPY_ARB, *
 * GL_STATIC_DRAW_ARB, GL_STATIC_READ_ARB, GL_STATIC_COPY_ARB,                  *
 * GL_DYNAMIC_DRAW_ARB, GL_DYNAMIC_READ_ARB, GL_DYNAMIC_COPY_ARB.               *
 ********************************************************************************/
GLuint createVBO(void* data, int dataSize, GLenum target, GLenum usage) {
  GLuint id = 0;  // 0 is reserved, glGenBuffersARB() will return non-zero id if success

  glGenBuffersARB(1, &id);                        // create a vbo
  glBindBufferARB(target, id);                    // activate vbo id to use
  glBufferDataARB(target, dataSize, data, usage); // upload data to video card

  // check data size in VBO is same as input array, if not return 0 and delete VBO
  int bufferSize = 0;
  glGetBufferParameterivARB(target, GL_BUFFER_SIZE_ARB, &bufferSize);
  if(dataSize != bufferSize)
  {
    glDeleteBuffersARB(1, &id);
    id = 0;
    fprintf(stderr,"[createVBO()] Data size is mismatch with input array\n");
  }

  return id;      // return VBO id
}
void deleteVBO(const GLuint vboId) {
  glDeleteBuffersARB(1, &vboId);
}

GLboolean isExtensionSupported(const char *extension) {
  const GLubyte *extensions = NULL;
  const GLubyte *version = NULL;
  GLubyte *where;
  /* Extension names should not have spaces. */
  where = (GLubyte *) strchr(extension, ' ');
  if (where || *extension == '\0')
    return GL_FALSE;
  version = glGetString(GL_VERSION);
  extensions = glGetString(GL_EXTENSIONS);
  return gluCheckExtension((const GLubyte *)extension, extensions);
}

/* Mouse Callbacks for panning and zooming. */
void mouseCB(int button, int state, int x, int y) {
  mouseX = x;
  mouseY = y;

  if(button == GLUT_LEFT_BUTTON)
  {
    if(state == GLUT_DOWN)
    {
      mouseLeftDown = GL_TRUE;
    }
    else if(state == GLUT_UP)
      mouseLeftDown = GL_FALSE;
  }

  else if(button == GLUT_RIGHT_BUTTON)
  {
    if(state == GLUT_DOWN)
    {
      mouseRightDown = GL_TRUE;
    }
    else if(state == GLUT_UP)
      mouseRightDown = GL_FALSE;
  }
}

void mouseMotionCB(int x, int y) {
  if(mouseLeftDown)
  {
    cameraAngleY += (x - mouseX);
    cameraAngleX += (y - mouseY);
    mouseX = x;
    mouseY = y;
    glutPostRedisplay();
  }
  if(mouseRightDown)
  {
    cameraDistance += (y - mouseY) * 0.2f;
    mouseY = y;
    glutPostRedisplay();
  }
}

void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ) {
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(posX, posY, posZ, targetX, targetY, targetZ, 0, 1, 0); // eye(x,y,z), focal(x,y,z), up(x,y,z)
}
